class
	STRING_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			visit_assignment_expression,
			visit_function_call,
			visit_function_declaration,
			visit_generic_expression,
			visit_jump_statement,
			visit_string_constant,
			end_visit_assignment_expression,
			end_visit_cast_expression,
			end_visit_function_call,
			end_visit_postfix_expression,
			end_visit_variable_declaration
		end
	PLATFORM
		export
			{NONE}all
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			create analyzer.make
			transformer.tree.accept (analyzer)
			analyzer.finalize
			--log (analyzer.get_analysis)
			log (analyzer.dependency_graph.get_statistics)
			transform_declarations (analyzer.supported_types)

			if has_transformed then
				transformer.tree.accept (Current)
			end
			Result := True
		end

	description: STRING
		do
			Result := "Transforming char arrays to strings"
		end

feature {NONE}

	analyzer: AST_STRING_ANALYZER

	expect_eiffel_string: BOOLEAN

	current_function_declaration: AST_FUNCTION_DECLARATION

feature {ANY}

	visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION): BOOLEAN
		local
			value: INTEGER
			one: AST_DECIMAL_CONSTANT
			idx: AST_EXPRESSION
		do
			if
				attached {AST_POSTFIX_EXPRESSION} node.target as target_expr and then
				attached {AST_VARIABLE_EXPRESSION} target_expr.main_expression as var_expr and then
				attached {AST_EIFFEL_STRING_TYPE} var_expr.type_binding
			then
				-- setting a character in an Eiffel string
				node.source.accept (Current)
				target_expr.main_expression.accept (Current)

				-- increment index
				if attached {AST_DECIMAL_CONSTANT} target_expr.array_expression as const then
					value := const.value.to_integer
					idx := create {AST_DECIMAL_CONSTANT}.make (value + 1, const.type_binding.ast_twin)
				else
					one := create {AST_DECIMAL_CONSTANT}.make (1, create {AST_BASIC_TYPE}.make (ast_bt_int))
					idx := create {AST_GENERIC_EXPRESSION}.make (c_op_plus, target_expr.array_expression.ast_twin, one)
				end

				-- replace assignment: `s[i] := source' -> `s.put_code(source, i+1)'
				replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed(<< target_expr.main_expression.ast_twin, ".put_code(", node.source.ast_twin, ".to_natural_32, ", idx, ")" >>, create {AST_BASIC_TYPE}.make (ast_bt_void)))
				Result := False
			else
				Result := True
			end

		end

	visit_function_call (node: AST_FUNCTION_CALL): BOOLEAN
		local
			decl: AST_DECLARATION
			extern: BOOLEAN
			actual_params: LIST [AST_EXPRESSION]
			declared_params: LIST [AST_NODE]
			eiffel_string_backup: BOOLEAN
		do
			eiffel_string_backup := expect_eiffel_string

			decl := node.declaration_binding
			if attached {AST_FUNCTION_DECLARATION} decl as f_decl then
				declared_params := f_decl.parameter_list.parameters
				extern := f_decl.is_extern
			else
				-- not supported
			end

			if decl /= Void and then not extern then
				-- set is_eiffel_string to True for AST_STRING_CONSTANTS where the called feature expects an Eiffel string
				from
					actual_params := node.parameters
					actual_params.start
					declared_params.start
				until
					actual_params.after or declared_params.after
				loop

					if
						(attached {AST_VARIABLE_DECLARATION} declared_params.item as param_decl and then
						attached {AST_EIFFEL_STRING_TYPE} param_decl.type) or
						attached {AST_EIFFEL_STRING_TYPE} declared_params.item
					then
						expect_eiffel_string := True
					else
						expect_eiffel_string := False
					end

					actual_params.item.accept (Current)

					actual_params.forth
					declared_params.forth
				end
				Result := False
			else
				Result := True
			end

			expect_eiffel_string := eiffel_string_backup
		end

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN
		do
			current_function_declaration := node
			Result := True
		end

	visit_generic_expression (node:AST_GENERIC_EXPRESSION): BOOLEAN
		do
			if
				analyzer.is_equality_check (node) and then
				attached {AST_CAST_EXPRESSION} node.left as left_cast and then
				attached {AST_CAST_EXPRESSION} node.right as right_cast and then
				attached {AST_VARIABLE_EXPRESSION} left_cast.expression.non_parentheses_expression as left_var and then
				attached {AST_VARIABLE_EXPRESSION} right_cast.expression.non_parentheses_expression as right_var and then
				attached {AST_EIFFEL_STRING_TYPE} left_var.type_binding and then
				attached {AST_EIFFEL_STRING_TYPE} right_var.type_binding
			then
				replace_node (node, create {AST_GENERIC_EXPRESSION}.make (node.type, left_var.ast_twin, right_var.ast_twin))
				Result := False
			else
				Result := True
			end
		end

	visit_jump_statement (node: AST_JUMP_STATEMENT): BOOLEAN
		do
			if node.type = c_kw_return and attached {AST_EIFFEL_STRING_TYPE} current_function_declaration.return_type then
				-- we return an Eiffel string
				expect_eiffel_string := True
				node.return_expression.accept (Current)
				expect_eiffel_string := False
				Result := False
			else
				Result := True
			end
		end

	visit_string_constant (node: AST_STRING_CONSTANT): BOOLEAN
		do
			if expect_eiffel_string then
				node.set_is_eiffel_string (True)
			end
		end

	end_visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION)
		do
			if
				attached {AST_VARIABLE_EXPRESSION} node.target as var and then
				attached {AST_EIFFEL_STRING_TYPE} var.type_binding and then
				attached {AST_STRING_CONSTANT} node.source as const
			then
				-- treat string constant as Eiffel string
				const.set_is_eiffel_string (True)
			end
		end

	end_visit_cast_expression (node: AST_CAST_EXPRESSION)
		local
			void_expr: AST_EIFFEL_EXPRESSION
		do
			if
				attached {AST_VARIABLE_EXPRESSION} node.expression as var_expr and then
				attached {AST_EIFFEL_STRING_TYPE} var_expr.type_binding and then
				attached {AST_GENERIC_EXPRESSION} node.parent as gen_expr and then
				analyzer.is_null_check (gen_expr)
			then
				create void_expr.make ("Void", create {AST_EIFFEL_STRING_TYPE}.make_string)
				replace_node (node.parent, create {AST_GENERIC_EXPRESSION}.make (gen_expr.type, var_expr.ast_twin, void_expr))
			end
		end

	end_visit_function_call (node: AST_FUNCTION_CALL)
		do
			if
				node.declaration_binding /= Void and then
				node.declaration_binding.is_extern
			then
				if  analyzer.string_h_functions.has (node.identifier_str) then
					handle_string_h_function_call (node)
				elseif node.identifier_str.is_equal ("sprintf") then
					handle_sprintf_function_call (node)
				end

			end
		end

	end_visit_postfix_expression (node: AST_POSTFIX_EXPRESSION)
		do
			if
				node.type = c_op_open_square_bracket and then
				attached {AST_EIFFEL_STRING_TYPE} node.main_expression.type_binding
			then
				log ("not supported! (should not happen)")
			end
		end

	end_visit_variable_declaration (node: AST_VARIABLE_DECLARATION)
		do
			if attached {AST_EIFFEL_STRING_TYPE} node.type and then node.initializer /= Void then
				if attached {AST_STRING_CONSTANT} node.initializer as const then
					const.set_is_eiffel_string (True)
				elseif
					attached {AST_CAST_EXPRESSION} node.initializer as cast_expr and then
					analyzer.is_null_pointer (cast_expr)
				then
					-- Eiffel does initialize `STRING' by default with `Void'
					delete_node (node.initializer)
				end
			end
		end

feature {NONE}

	has_transformed: BOOLEAN

	transform_declarations (types: LIST [AST_TYPE])
		local
			type_binding_updater: AST_AGENT_VISITOR
		do
			across
				types as c
			loop
				if
					attached {AST_ARRAY_TYPE} c.item as array_type and then
					array_type.dimension /= Void
				then
					replace_node (c.item, create {AST_EIFFEL_STRING_TYPE}.make_with_dimension (array_type.dimension.ast_twin))
				else
					replace_node (c.item, create {AST_EIFFEL_STRING_TYPE}.make_string)
				end
			end

			has_transformed := not types.is_empty

			if has_transformed then
				-- adjust type bindings
				create type_binding_updater
				type_binding_updater.set_visit_function_call_agent (agent update_function_call_type_binding)
				transformer.tree.accept (type_binding_updater)
			end
		end

	handle_string_h_function_call (node: AST_FUNCTION_CALL)
		local
			p, p1, p2, p3: AST_EXPRESSION
			replaceable: BOOLEAN
			is_arg1_eif_string: BOOLEAN
			as_stmt: BOOLEAN
			name: STRING
		do
			name := node.identifier_str
			p := node.parameters.first
			if attached {AST_PARENTHESIZED_EXPRESSION} p as expr then
				p1 := p.non_parentheses_expression.ast_twin
			else
				p1 := p.ast_twin
			end

			as_stmt := attached {AST_EXPRESSION_STATEMENT} node.parent
			replaceable := (analyzer.replaceable_string_h_functions_as_statement.has (name) and then as_stmt) or analyzer.replaceable_string_h_functions_as_expression.has (name)
			is_arg1_eif_string := attached {AST_EIFFEL_STRING_TYPE} p1.type_binding

			if
				 replaceable and then is_arg1_eif_string
			then
				-- replace the C function with an Eiffel method

				if node.parameters.count > 1 then
					p2 := node.parameters.at (2).ast_twin

					if attached {AST_STRING_CONSTANT} p2 as const then
						const.set_is_eiffel_string (True)
					elseif not attached {AST_EIFFEL_STRING_TYPE} p2.type_binding then
						-- convert second argument to an Eiffel string
						p2 := create {AST_EIFFEL_EXPRESSION}.make_mixed (<< "eif_string (", p2, ")" >>, create {AST_EIFFEL_STRING_TYPE}.make_string)
					end
				end

				inspect node.parameters.count
				when 1 then
					if node.identifier_str.is_equal ("strlen") then
						replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< p1, ".count.to_natural_32" >>, create {AST_BASIC_TYPE}.make (ast_bt_unsigned_int)))
					end
				when 2 then
					if node.identifier_str.is_equal ("strcpy") then
						if attached {AST_STRING_CONSTANT} p2 as const then
							replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< p1, " := ", p2 >>, create {AST_BASIC_TYPE}.make (ast_bt_void)))
						else
							replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< p1, " := ", p2, ".twin" >>, create {AST_BASIC_TYPE}.make (ast_bt_void)))
						end
					end
					if node.identifier_str.is_equal ("strcmp") then
						replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< p1, ".three_way_comparison (", p2, ")" >>, create {AST_BASIC_TYPE}.make (ast_bt_int)))
					end
					if node.identifier_str.is_equal ("strcat") then
						replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< p1, ".append (", p2, ")" >>, create {AST_BASIC_TYPE}.make (ast_bt_int)))
					end
				when 3 then
					p3 := node.parameters.last.ast_twin
					if attached {AST_DECIMAL_CONSTANT} p3 as const then
						p3 := create {AST_DECIMAL_CONSTANT}.make (const.value.to_integer, create {AST_BASIC_TYPE}.make (ast_bt_int))
					else
						p3 := create {AST_EIFFEL_EXPRESSION}.make_mixed (<< p3, ".as_integer_32" >>, create {AST_BASIC_TYPE}.make (ast_bt_int))
					end

					if node.identifier_str.is_equal ("strncat") then
						replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< p1, ".append (", p2, ".substring (1, ", p2.ast_twin, ".count.min (", p3, ")))" >>, create {AST_BASIC_TYPE}.make (ast_bt_int)))
					end
					if node.identifier_str.is_equal ("strncmp") then
						replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< p1, ".substring (1, ", p1.ast_twin, ".count.min (", p3, ")).three_way_comparison (", p2, ".substring (1, ", p2.ast_twin, ".count.min (", p3, ")))" >>, create {AST_BASIC_TYPE}.make (ast_bt_int)))
					end
				else
					-- not supported
				end
			else
				-- no replace possible -> convert Eiffel strings in parameter list to C strings
				across
					node.parameters as param
				loop
					p := param.item
					if attached {AST_EIFFEL_STRING_TYPE} p.type_binding then
						replace_node (p, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< "ce_string (", p.ast_twin, ")" >>, create {AST_POINTER_TYPE}.make (create {AST_BASIC_TYPE}.make (ast_bt_char))))
					end
				end
			end
		end

	handle_sprintf_function_call (node: AST_FUNCTION_CALL)
		local
			helper: PRINTF_HELPER
			target_expr: AST_EXPRESSION
			eiffel_parameters: AST_EXPRESSION
		do
			create helper
			node.parameters.start
			target_expr := node.parameters.item
			node.parameters.forth

			if
				attached {AST_VARIABLE_EXPRESSION} target_expr.non_parentheses_expression as target_var and then
				attached {AST_EIFFEL_STRING_TYPE} target_var.type_binding
			then
				eiffel_parameters := helper.create_eiffel_parameters (node.parameters)

				replace_node (node, create {AST_ASSIGNMENT_EXPRESSION}.make (target_var.ast_twin, eiffel_parameters))
			else
				across
					node.parameters as p
				loop
					if attached {AST_EIFFEL_STRING_TYPE} p.item.type_binding then
						-- convert eiffel strings to ce_strings for external sprintf call
						replace_node (p.item, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< "ce_string (", p.item.ast_twin, ")" >>, create {AST_ARRAY_TYPE}.make (create {AST_BASIC_TYPE}.make (Ast_bt_char), Void)))
					elseif attached {AST_STRING_CONSTANT} p.item as const and then const.is_eiffel_string then
						-- make sure `is_eiffel_string' is set to `False'
						const.set_is_eiffel_string (False)
					end
				end
			end

		end

	update_function_call_type_binding (node: AST_FUNCTION_CALL): BOOLEAN
		do
			if
				attached {AST_FUNCTION_DECLARATION} node.declaration_binding as func_decl and then
				attached {AST_EIFFEL_STRING_TYPE} func_decl.return_type
			then
				node.type_binding := func_decl.return_type.ast_twin
			end

			Result := True
		end

end
