class
	PRINTF_HELPER
inherit
	AST_UTILITIES


feature {ANY}

	is_printf_replaceable (c_params: LIST [AST_EXPRESSION]): BOOLEAN
		local
			format_string: STRING
			orig_pos: CURSOR
		do
			if attached {AST_STRING_CONSTANT} c_params.item as string_const then
				orig_pos := c_params.cursor
				format_string := string_const.value
				c_params.forth

				Result := check_print_statement (format_string, c_params)

				-- restore cursor position
				c_params.go_to (orig_pos)
			elseif c_params.islast then
				Result := True
			else
				-- not supported
				Result := False
			end

		end

	create_eiffel_parameters (c_params: LIST [AST_EXPRESSION]): AST_EXPRESSION
		require
			is_printf_replaceable (c_params)
		local
			format_string: STRING
		do
			if attached {AST_STRING_CONSTANT} c_params.item as string_const then
				format_string := string_const.value
				c_params.forth

				Result := parse_print_statement (format_string, c_params)
			elseif c_params.islast then
				if attached {AST_EIFFEL_STRING_TYPE} c_params.item.type_binding then
					Result := c_params.item.ast_twin
				else
					Result := create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<<"eif_string (", create_eif_string_argument (c_params.item.ast_twin), ")">>,
							create {AST_EIFFEL_STRING_TYPE}.make_string
						)
				end
			end
		end

feature {NONE} -- Implementation

	check_print_statement (format_string: STRING; c_params: LIST [AST_EXPRESSION]): BOOLEAN
			-- returns if the printf statement is replaceable
		local
			format: STRING
		do
			Result := True

			across
				format_string as c
			until
				Result = False
			loop
				if c.item = '%%' then
					c.forth
					if not (c.item = '%%') then
						-- build format string
						create format.make_empty
						from
						until
							is_specifier (c.item)
						loop
							format.append_character (c.item)
							c.forth
						end
						if is_supported (c.item, format) then
							Result := check_supported_type (c.item, format, c_params)
						else
							Result := False
						end
					end
				end
			end

			if not c_params.after then
				Result := False
			end
		end

	check_supported_type (type: CHARACTER; format: STRING; c_params: LIST [AST_EXPRESSION]): BOOLEAN
		local
			supported_types: ARRAYED_LIST [INTEGER]
			format_args: LIST [STRING]
		do

			-- parse format string for width and precision
			-- the format is [width][.precision] where both parts are optional
			format_args := format.split ('.')
			-- if there is a width, it is in format_args.at (1)
			-- if there is a precision, it is in format_args.at (2)
			if
				not format_args.at (1).is_empty and then
				format_args.at (1).is_equal ("*")
			then
				c_params.forth
			end
			if
				format_args.count = 2 and then
				not c_params.after and then
				format_args.at (2).is_equal ("*")
			then
				c_params.forth
			end

			-- check type of parameters
			inspect type
			when 'c' then
				create supported_types.make_from_array (<< ast_bt_char, ast_bt_unsigned_char, ast_bt_int, ast_bt_unsigned_int, ast_bt_short, ast_bt_unsigned_short >>)
				if
					attached {AST_DECIMAL_CONSTANT} c_params.item or else
					(attached {AST_BASIC_TYPE} c_params.item.type_binding as basic_type and then
					supported_types.has (basic_type.type))
				then
					Result := True
				end
			when 'd','i' then
				create supported_types.make_from_array (<< ast_bt_char, ast_bt_unsigned_char, ast_bt_short, ast_bt_unsigned_short, ast_bt_int, ast_bt_long, ast_bt_unsigned_int, ast_bt_unsigned_long, ast_bt_long_long, ast_bt_unsigned_long_long >>)
				if
					attached {AST_BASIC_TYPE} c_params.item.type_binding as basic_type and then
					supported_types.has (basic_type.type)
				then
					Result := True
				end
			when 'f' then
				create supported_types.make_from_array (<< ast_bt_double, ast_bt_float >>)
				if
					attached {AST_BASIC_TYPE} c_params.item.type_binding as basic_type and then
					supported_types.has (basic_type.type)
				then
					Result := True
				end
			when 's' then
				-- every type is ok
				Result := True
			when 'u' then
				create supported_types.make_from_array (<< ast_bt_unsigned_char, ast_bt_unsigned_short, ast_bt_unsigned_int, ast_bt_unsigned_long, ast_bt_char, ast_bt_short, ast_bt_int, ast_bt_long, ast_bt_long_long, ast_bt_unsigned_long_long >>)
				if
					attached {AST_BASIC_TYPE} c_params.item.type_binding as basic_type and then
					supported_types.has (basic_type.type)
				then
					Result := True
				end
			end

			c_params.forth
		end

	parse_print_statement (format_string: STRING; c_params: LIST [AST_EXPRESSION]): AST_EXPRESSION
		local
			buffer: STRING
			format: STRING
		do
			create buffer.make_empty
			across
				format_string as c
			until
				buffer = Void
			loop
				if c.item = '%%' then
					c.forth
					if (c.item = '%%') then
						buffer.append ("%%%%")
					else
						-- build format string
						format := create {STRING}.make_empty
						from
						until
							is_specifier (c.item)
						loop
							format.append_character (c.item)
							c.forth
						end
						Result := handle_supported_type (c.item, format, Result, buffer, c_params)
					end
				else
					buffer.append_character (c.item)
				end
			end
			if buffer = Void or not c_params.after then
				Result := Void
			elseif not buffer.is_empty then
				Result := build_concatenation (Result, create_eiffel_string (buffer))
			end

		end

	is_specifier (c: CHARACTER): BOOLEAN
		do
			inspect c
			when 'c', 'd', 'i', 'e', 'E', 'f', 'g', 'G', 'o', 's', 'u', 'x', 'X', 'p', 'n' then
				Result := True
			else
				Result := False
			end
		end

	is_supported (type: CHARACTER; format: STRING): BOOLEAN
		require
			format /= Void
		local
			c: CHARACTER
		do
			-- check for supported types
			inspect type
			when 'c', 'd', 'i', 'f', 's', 'u' then
				Result := True
			else
				Result := False
			end

			-- check format parameters
			if not format.is_empty and Result then
				-- no flags supported
				inspect format.at (1)
				when '-', '+', ' ', '#', '0' then
					Result := False
				else
				end

				-- precision not supported for strings
				if Result and then type = 's' then
					if format.has ('.') then
						Result := False
					end
				end

				-- width not supported for strings / characters / unsigned integer
				if Result and (type = 's' or type = 'c' or type = 'u') then
					c := format.at (1)
					if c.is_digit or c = '*' then
						-- width specified for a string / character
						Result := False
					end
				end

				--precision not supported for integer
				if Result and (type = 'd' or type = 'i' or type = 'u') then
					if format.has ('.') then
						-- precision specified for integer type
						Result := FALSE
					end
				end

				-- length parameter is not supported
				if Result then
					inspect format.at (format.count)
					when 'h', 'l', 'L' then
						Result := False
					else
					end
				end
			end
		end

	handle_supported_type (type: CHARACTER; format: STRING; eiffel_parameter: AST_EXPRESSION; buffer: STRING; c_params: LIST [AST_EXPRESSION]): AST_EXPRESSION
		require
			buffer /= Void
			c_params /= Void
		local
			format_args: LIST [STRING]
			width, precision: INTEGER
			variable_width, variable_precision: AST_EXPRESSION
			has_variable_width, has_variable_precision: BOOLEAN

			conversion_string: STRING
			array: ARRAY[ANY]
			list: ARRAYED_LIST[ANY]
		do
			Result := eiffel_parameter
			if not buffer.is_empty then
				Result := build_concatenation (Result, create_eiffel_string (buffer))
				buffer.wipe_out
			end

			-- parse format string for width and precision
			-- the format is [width][.precision] where both parts are optional
			format_args := format.split ('.')
			-- if there is a width, it is in format_args.at (1)
			-- if there is a precision, it is in format_args.at (2)
			if not format_args.at (1).is_empty then
				if format_args.at (1).is_equal ("*") then
					variable_width := c_params.item.ast_twin
					has_variable_width := True
					c_params.forth
				else
					width := format_args.at (1).to_integer
				end
			else
				-- set default value
				width := 1
			end
			if format_args.count = 2 and not c_params.after then
				if format_args.at (2).is_empty then
					--use 0 as precision
					precision := 0
				elseif format_args.at (2).is_equal ("*") then
					variable_precision := c_params.item.ast_twin
					has_variable_precision := True
					c_params.forth
				else
					precision := format_args.at (2).to_integer
				end
			else
				-- set default value for floats
				-- (only floats support precision)
				precision := 6
			end

			-- generate eiffel parameters
			inspect type
			when 'c' then
				-- precision has no effect for type 'c', so it is ignored here. width is not supported

				if attached {AST_CHARACTER_CONSTANT} c_params.item as constant then
					-- append it as string (remove '')
					Result := build_concatenation (Result, create_eiffel_string (constant.value.at (2).out))
				elseif attached {AST_DECIMAL_CONSTANT} c_params.item as constant then
					-- convert to 8 bit character (default in c)
					Result := build_concatenation (Result, create_eiffel_string (constant.value.to_integer.to_character_8.out))
				else
					if
						attached {AST_EIFFEL_EXPRESSION} c_params.item as expr and then
						attached {STRING} expr.mixed.at (2) as conversion and then
						conversion.ends_with (".code.to_integer_8")
					then
						conversion.replace_substring_all (".code.to_integer_8", ".out")
						Result := build_concatenation (Result, expr.ast_twin)
					else
						-- convert to 8-bit character
						Result := build_concatenation (Result, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< create {AST_PARENTHESIZED_EXPRESSION}.make (c_params.item.ast_twin), ".to_character_8.out" >>, create {AST_EIFFEL_STRING_TYPE}.make_string))
					end
				end
			when 'd','i' then
				-- precision is not supported for 'd' and 'i'

				if
					attached {AST_DECIMAL_CONSTANT} c_params.item as constant and then
					not has_variable_width
				then
					-- we know the width and the integer, so we can build the string at compile time
					Result := build_concatenation (Result, create_eiffel_string ((create {FORMAT_INTEGER}.make (width)).formatted (constant.value.out.to_integer)))
				elseif
					attached {AST_BASIC_TYPE} c_params.item.type_binding as basic_type
				then
					inspect basic_type.type
					when ast_bt_char, ast_bt_unsigned_char, ast_bt_short, ast_bt_unsigned_short, ast_bt_int, ast_bt_long then
						conversion_string := ""
					when ast_bt_unsigned_int, ast_bt_unsigned_long, ast_bt_long_long, ast_bt_unsigned_long_long then
						conversion_string := ".to_integer_32"
					end

					if has_variable_width then
						array := << "(create {FORMAT_INTEGER}.make (", variable_width, ")).formatted (", create {AST_PARENTHESIZED_EXPRESSION}.make (c_params.item.ast_twin), conversion_string + ")"  >>
					elseif width > 1 then
						array := << "(create {FORMAT_INTEGER}.make (" + width.out + ")).formatted (", create {AST_PARENTHESIZED_EXPRESSION}.make (c_params.item.ast_twin), conversion_string + ")"  >>
					else
						array := << create {AST_PARENTHESIZED_EXPRESSION}.make (c_params.item.ast_twin), conversion_string + ".out" >>
					end
					Result := build_concatenation (Result, create {AST_EIFFEL_EXPRESSION}.make_mixed (array, create {AST_EIFFEL_STRING_TYPE}.make_string))
				end
			when 'f' then
				-- width and precision is supported for 'f'
			   	if has_variable_precision and has_variable_width then
			   		array := (<< create {AST_PARENTHESIZED_EXPRESSION}.make (variable_width), ".max (", variable_precision.ast_twin, "), ", variable_precision >>)
			   	elseif has_variable_precision then
			   		array := (<< "(" +  width.out + ").max (", variable_precision.ast_twin, "), ", variable_precision >>)
			   	elseif has_variable_width then
			   		array := (<< create {AST_PARENTHESIZED_EXPRESSION}.make (variable_width), ".max (" + precision.out + "), " + precision.out >>)
			   	else
					width := width.max (precision)
					array := (<< width.out + ", " + precision.out >>)
			   	end

			   	create list.make_from_array (array)
			   	list.put_front ("(create {FORMAT_DOUBLE}.make (")
			   	list.extend (")).formatted (")
				list.extend (c_params.item.ast_twin)
				list.extend (")")

				Result := build_concatenation (Result, create {AST_EIFFEL_EXPRESSION}.make_mixed (list.to_array, create {AST_EIFFEL_STRING_TYPE}.make_string))
			when 's' then
				-- width and precision is not supported for 's'
				if attached {AST_STRING_CONSTANT} c_params.item as string_constant then
					string_constant.is_eiffel_string := True
					Result := build_concatenation (Result, string_constant.ast_twin)
				elseif attached {AST_EIFFEL_STRING_TYPE} c_params.item.type_binding then
					Result := build_concatenation (Result, c_params.item.ast_twin)
				else
					Result := build_concatenation (Result, create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<< "eif_string (", create_eif_string_argument (c_params.item.ast_twin), ")" >>,
							create {AST_EIFFEL_STRING_TYPE}.make_string
						))
				end
			when 'u' then
				-- width and precision is not supported for 'u'

				if attached {AST_DECIMAL_CONSTANT} c_params.item as constant then
					Result := build_concatenation (Result, create_eiffel_string (constant.value.out.to_integer.to_natural_32.out))
				elseif
					attached {AST_BASIC_TYPE} c_params.item.type_binding as basic_type
				then
					inspect basic_type.type
					when ast_bt_unsigned_char, ast_bt_unsigned_short, ast_bt_unsigned_int, ast_bt_unsigned_long then
						conversion_string := ""
					when ast_bt_char, ast_bt_short, ast_bt_int, ast_bt_long, ast_bt_long_long, ast_bt_unsigned_long_long then
						conversion_string := ".to_natural_32"
					end

					Result := build_concatenation (Result, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< create {AST_PARENTHESIZED_EXPRESSION}.make (c_params.item.ast_twin), conversion_string + ".out" >>, create {AST_EIFFEL_STRING_TYPE}.make_string))
				end
			end

			c_params.forth
		ensure
			buffer.is_empty
		end

	create_eiffel_string (string: STRING): AST_STRING_CONSTANT
		require
			string /= Void
		do
			create Result.make (string.twin)
			Result.is_eiffel_string := True
		end

	build_concatenation (orig, new: AST_EXPRESSION): AST_EXPRESSION
		require
			new /= Void
		do
			if orig = Void then
				Result := new
			elseif
				attached {AST_STRING_CONSTANT} orig as sc1 and
				attached {AST_STRING_CONSTANT} new as sc2
			then
				sc1.value.append (sc2.value)
				Result := orig
			elseif
				attached {AST_GENERIC_EXPRESSION} orig as ge and then
				attached {AST_STRING_CONSTANT} ge.right as sc1 and then
				attached {AST_STRING_CONSTANT} new as sc2
			then
				sc1.value.append (sc2.value)
				Result := orig
			else
				Result := create {AST_EIFFEL_EXPRESSION}.make_mixed (<< orig, " + ", new >>, orig.type_binding)
			end
		ensure
			Result /= Void
		end

	create_eif_string_argument (node: AST_EXPRESSION): AST_EXPRESSION
			-- Adds a cast to `char *' if the given expression has another type
		require
			node /= Void and then node.parent = Void
		do
			if attached {AST_POINTER_TYPE} node.type_binding or attached {AST_ARRAY_TYPE} node.type_binding then
				if attached {AST_BASIC_TYPE} deref (node.type_binding) as bt and then bt.type = Ast_bt_char then
					Result := node
				else
					Result := create {AST_EIFFEL_EXPRESSION}.make_mixed (<<
							"create {CE_POINTER [INTEGER_8]}.make_shared (",
							create {AST_PARENTHESIZED_EXPRESSION}.make (node),
							".c_pointer)"
						>>, create {AST_POINTER_TYPE}.make (create {AST_BASIC_TYPE}.make (Ast_bt_char)))
				end
			else
				(create {EXCEPTIONS}).raise ("Doesn't have type %"string%": " + node.c_code)
			end
		ensure
			Result /= Void and then Result.parent = Void
		end


end
