class
	AST_STRING_ANALYZER

inherit
	ANALYZER
		redefine
			make,
			collector_visit_type,
			visit_array_type,
			visit_assignment_expression,
			visit_cast_expression,
			visit_eiffel_expression,
			visit_function_call,
			visit_generic_expression,
			visit_initializer,
			visit_jump_statement,
			visit_pointer_type,
			visit_postfix_expression,
			visit_unary_expression,
			visit_variable_declaration,
			visit_variable_expression
		end
	VIEW_INTERFACE

create
	make

feature -- Initialization

	make
		do
			Precursor

			create dependency_graph.make
		end

feature {STRING_TRANSFORMER}

	string_h_functions: ARRAYED_LIST [STRING]
			-- all considered functions from `string.h'
		once
			create Result.make_from_array (<< "strcat", "strncat", "strchr", "strcmp", "strncmp", "strcoll", "strcpy", "strncpy", "strcspn", "strerror", "strlen", "strpbrk", "strrchr", "strspn", "strstr", "strtok", "strxfrm" >>)
			Result.compare_objects
		end

	replaceable_string_h_functions_as_expression: ARRAYED_LIST [STRING]
			-- those functions can be replaced in expressions
		once
			create Result.make_from_array (<< "strcmp", "strncmp", "strlen" >>)
			Result.compare_objects
		end

	replaceable_string_h_functions_as_statement: ARRAYED_LIST [STRING]
			-- those functions can be replaced but only as statements
		once
			create Result.make_from_array (<< "strcat", "strcpy", "strncat" >>)
			Result.compare_objects
		end

	all_args_supportable_string_h_functions: ARRAYED_LIST [STRING]
			-- for those functions all arguments can be an Eiffel string (wrapped with `ce_string()')
		once
			create Result.make_from_array (<< "strcmp", "strcoll", "strcspn", "strlen", "strncmp", "strspn" >>)
			Result.compare_objects
		end

	snd_arg_supportable_string_h_functions: ARRAYED_LIST [STRING]
			-- for those functions only the second argument can be an Eiffel string (wrapped with `ce_string()')
		once
			create Result.make_from_array (<< "strcat", "strcpy", "strncat", "strncpy", "strpbrk", "strstr", "strtok", "strxfrm" >>)
			Result.compare_objects
		end

	printf_functions: ARRAYED_LIST [STRING]
			--  all considered printf functions
		once
			create Result.make_from_array (<< "printf", "fprintf", "sprintf" >>)
			Result.compare_objects
		end

feature {STRING_TRANSFORMER}

	supported_types: LIST [AST_TYPE]
		require
			dependency_graph.is_finalized
		do
			Result := dependency_graph.good_items
		end

feature -- Access

	dependency_graph:  DEPENDENCY_GRAPH [AST_TYPE]

feature -- Basic operations

	finalize
		do
			dependency_graph.finalize
		end

	get_analysis: STRING
		do
			Result :="*** Dependencygraph ***%N"
			Result.append (dependency_graph.to_string)
		end


feature {ANY}

	visit_array_type (node: AST_ARRAY_TYPE): BOOLEAN
		do
			if is_extern_function_declaration then
				Result := False
			elseif is_string_type (node) then
				insert_type (node)
			end

			-- do not visit subtype
			Result := False
		end

	visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION): BOOLEAN
		local
			target_decl: AST_VARIABLE_DECLARATION
			source_list: LIST [AST_TYPE]
		do
			if node.type = c_op_assign and then is_string_type (node.type_binding) then
				source_list := collect_in_sub_tree (node.source)

				if attached {AST_VARIABLE_EXPRESSION} node.target as target_expr then
					target_decl ?= target_expr.declaration_binding
				elseif attached {AST_POSTFIX_EXPRESSION} node.target as target_expr then
					target_decl ?= target_expr.declaration_binding
				end

				if
					(attached {AST_POSTFIX_EXPRESSION} node.target as expr and then
					expr.type = C_op_open_square_bracket) or
					attached {AST_UNARY_EXPRESSION} node.target or
					(attached {AST_POSTFIX_EXPRESSION} node.source as expr and then
					expr.type = C_op_open_square_bracket) or
					attached {AST_UNARY_EXPRESSION} node.source
				then
					-- unsupported assignment
					dependency_graph.mark_all_bad (collect_in_sub_tree (node.target), Assignment_flag)
					dependency_graph.mark_all_bad (source_list, Assignment_flag)
				elseif
					source_list.is_empty and then
					not (attached {AST_CAST_EXPRESSION} node.source as cast and then
					is_null_pointer (cast)) and then
					not attached {AST_STRING_CONSTANT} node.source
				then
					-- unsupported assignment
					dependency_graph.mark_all_bad (collect_in_sub_tree (node.target), Assignment_flag)
				elseif
					target_decl /= Void and then
					is_string_type (node.source.type_binding)
				then
					if attached {AST_VARIABLE_EXPRESSION} node.source as source and then source.declaration_binding /= Void then
						dependency_graph.insert_dependency (source.declaration_binding.type, target_decl.type)
					elseif attached {AST_FUNCTION_CALL} node.source as source and then source.declaration_binding /= Void then
						dependency_graph.insert_dependency (source.declaration_binding.return_type, target_decl.type)
					else
						across
							source_list as c
						loop
							if c.item /= target_decl.type then
								dependency_graph.insert_dependency (c.item, target_decl.type)
							end
						end
					end
				end
			end
			Result := True
		end

	visit_cast_expression (node: AST_CAST_EXPRESSION): BOOLEAN
		local
			is_target_string_type, is_source_string_type: BOOLEAN
		do
			is_target_string_type := is_string_type (node.type)
			is_source_string_type := is_string_type (node.expression.type_binding)

			if is_null_pointer (node) then
				Result := False
			elseif is_source_string_type and then not is_target_string_type then
				-- not supported cast
				dependency_graph.mark_bad (node.expression.type_binding, Cast_flag)
				Result := True
			elseif is_target_string_type and then not is_source_string_type then
				insert_type (node.type)
				dependency_graph.mark_bad (node.type, Cast_flag)
				Result := True
			else
				Result := True
			end
		end

	visit_eiffel_expression (node: AST_EIFFEL_EXPRESSION): BOOLEAN
		do
			-- no clue what happens inside an eiffel_expression so we can't translate any strings used inside.
			dependency_graph.mark_all_bad (collect_in_list (node.mixed), Other_flag)

			Result := False
		end

	visit_function_call (node: AST_FUNCTION_CALL): BOOLEAN
		local
			decl: AST_DECLARATION
			return_type: AST_TYPE
			extern: BOOLEAN
			var_arg: BOOLEAN
			string_types: LIST [AST_TYPE]
			actual_params: LIST [AST_EXPRESSION]
			declared_params: LIST [AST_NODE]
			decl_type: AST_TYPE
		do
			decl := node.declaration_binding
			if attached {AST_FUNCTION_DECLARATION} decl as f_decl then
				return_type := f_decl.return_type
				declared_params := f_decl.parameter_list.parameters
				extern := f_decl.is_extern
				var_arg := f_decl.parameter_list.is_vararg
			else
				-- not supported
			end

			if decl /= Void and then extern and then string_h_functions.has (decl.identifier) then
				handle_string_h_function_call (node)
				Result := True
			elseif decl /= Void and then extern and then printf_functions.has (decl.identifier) then
				handle_printf_function_call (node)
				Result := True
			elseif decl /= Void and not extern and not var_arg then
				-- connect actual parameters to declared parameters
				from
					actual_params := node.parameters
					actual_params.start
					declared_params.start
				until
					actual_params.after or declared_params.after
				loop
					if attached {AST_VARIABLE_DECLARATION} declared_params.item as param_decl then
						decl_type := param_decl.type
					elseif attached {AST_TYPE} declared_params.item then
						decl_type ?= declared_params.item
					end

					if
						decl_type /= Void and then
						is_string_type (decl_type)
					then
						if
							attached {AST_UNARY_EXPRESSION} actual_params.item as unary and then
							unary.type = C_op_bitand and then
							attached {AST_BASIC_TYPE} unary.expression.type_binding as bt and then
							bt.type = Ast_bt_char
						then
							-- actual argument is like `& c' with c a character
							dependency_graph.mark_bad (decl_type, Pointer_flag)
						else
							insert_dependency (actual_params.item, decl_type)

							-- visit actual parameter
							actual_params.item.accept (Current)
						end
					else
						-- visit actual parameter
						actual_params.item.accept (Current)
					end

					actual_params.forth
					declared_params.forth
				end
				Result := False
			else
				--not supported -> mark all used string types bad

				string_types := collect_in_list (node.parameters)
				if is_string_type (return_type) then
					string_types.extend (return_type)
				end

				dependency_graph.mark_all_bad (string_types, External_function_flag)
				Result := True
			end
		end

	visit_generic_expression (node:AST_GENERIC_EXPRESSION): BOOLEAN
		local
			left_type, right_type: AST_TYPE
		do
			if is_null_check (node) then
				Result := False
			elseif
				is_equality_check (node) and then
				attached {AST_CAST_EXPRESSION} node.left as left_cast and then
				attached {AST_CAST_EXPRESSION} node.right as right_cast and then
				attached {AST_VARIABLE_EXPRESSION} left_cast.expression.non_parentheses_expression as left_var and then
				attached {AST_VARIABLE_EXPRESSION} right_cast.expression.non_parentheses_expression as right_var and then
				is_string_type (left_var.type_binding) and then
				is_string_type (right_var.type_binding)
			then
				left_type := left_var.type_binding
				right_type := right_var.type_binding
				if left_type /= right_type then
					dependency_graph.insert_dependency (left_type, right_type)
				end
				Result := False
			else
				Result := True
			end
		end

	visit_initializer (node: AST_INITIALIZER): BOOLEAN
		do
			if not is_string_type (node.type_binding) then
				dependency_graph.mark_all_bad (collect_in_list (node.elements), Initializer_flag)
				Result := True
			end
		end

	visit_jump_statement (node: AST_JUMP_STATEMENT): BOOLEAN
		do
			if node.type = c_kw_return and is_string_type (current_function_declaration.return_type) then
				insert_dependency (node.return_expression, current_function_declaration.return_type)
			end

			Result := True
		end

	visit_pointer_type (node: AST_POINTER_TYPE): BOOLEAN
		do
			if is_extern_function_declaration then
				Result := False
			elseif is_string_type (node) then
				insert_type (node)
			end

			-- do not visit subtype
			Result := False
		end

	visit_postfix_expression (node: AST_POSTFIX_EXPRESSION): BOOLEAN
		do
			if
				node.type = c_op_open_square_bracket and then
				is_string_type (node.main_expression.type_binding)
			then
				-- array access like `s[3]' are not supported.
				dependency_graph.mark_all_bad (collect_in_sub_tree (node.main_expression), Array_flag)
			end
		end

	visit_unary_expression (node: AST_UNARY_EXPRESSION): BOOLEAN
		do
			if node.type = C_op_bitand then
				dependency_graph.mark_all_bad (collect_in_sub_tree (node.expression), Pointer_flag)
			end
			Result := True
		end

	visit_variable_declaration (node: AST_VARIABLE_DECLARATION): BOOLEAN
		do
			if is_extern_function_declaration then
				-- ignore declarations
				Result := False
			elseif is_string_type (node.type) then
				node.type.accept (Current)

				if node.is_extern then
					dependency_graph.mark_bad (node.type, External_flag)
				elseif attached {AST_STRUCT_DECLARATION} node.parent then
					dependency_graph.mark_bad (node.type, Struct_member_flag)
				end

				if node.initializer /= Void then
					-- visit initializer
					node.initializer.accept (Current)

					if attached {AST_INITIALIZER} node.initializer then
						dependency_graph.mark_bad (node.type, Initializer_flag)
					elseif
						attached {AST_UNARY_EXPRESSION} node.initializer.non_parentheses_expression as unary and then
						unary.type = C_op_bitand and then
						attached {AST_BASIC_TYPE} unary.expression.type_binding as bt and then
						bt.type = Ast_bt_char
					then
						-- initializer is like `& c' with c a character
						dependency_graph.mark_bad (node.type, Pointer_flag)
					else
						-- connect initializer to declaration
						insert_dependency (node.initializer, node.type)
					end
				end

				Result := False
			else
				Result := True
			end
		end

	visit_variable_expression (node: AST_VARIABLE_EXPRESSION): BOOLEAN
		do
			if
				attached {AST_VARIABLE_DECLARATION} node.declaration_binding as decl and then
				is_string_type (decl.type)
			then
				-- check for pointer arithmetik
				if
					attached {AST_GENERIC_EXPRESSION} node.non_parentheses_parent as gen_expr and then
					gen_expr.is_additive_expression
				then
					dependency_graph.mark_bad (decl.type, Pointer_flag)
				end
			elseif
				attached {AST_UNARY_EXPRESSION} node.non_parentheses_parent as unary_expr and then
				unary_expr.type = c_op_bitand and then
				attached {AST_FUNCTION_DECLARATION} node.declaration_binding as func_decl
			then
				-- function pointers are not supported
				if is_string_type (func_decl.return_type) then
					dependency_graph.mark_bad (func_decl.return_type, Function_pointer_flag)
				end

				dependency_graph.mark_all_bad (collect_in_list (func_decl.parameter_list.parameters), Function_pointer_flag)
			end
			Result := True
		end

feature {NONE} -- ANALYZER

	collector_visit_type (type: AST_TYPE): BOOLEAN
		do
			if collect_type (type) then
				collected_types.extend (type)
			end

			-- don't visit subtype
			Result := False
		end

feature {NONE} -- Implementation

	collect_type (type: AST_TYPE): BOOLEAN
		do
			Result := is_string_type (type)
		end

	insert_dependency (node: AST_NODE; type: AST_TYPE)
		local
			types: LIST [AST_TYPE]
		do
			types := collect_in_sub_tree (node)
			if types.is_empty then
				-- at least insert type
				insert_type (type)
			else
				across
					types as c
				loop
					if c.item /= type then
						dependency_graph.insert_dependency (c.item, type)
					end
				end
			end
		end

	insert_type (node: AST_TYPE)
		local
			is_extern: BOOLEAN
			identifier: STRING
			flags: INTEGER
		do
			flags := 0

			if attached {AST_DECLARATION} node.parent as decl then
				if attached {AST_FUNCTION_DECLARATION} decl as func_decl then
					is_extern := func_decl.is_extern
					identifier := "return value of " + func_decl.identifier
				elseif attached {AST_VARIABLE_DECLARATION} decl as var_decl then
					is_extern := var_decl.is_extern
					identifier := var_decl.identifier.string
					if attached {AST_STRUCT_DECLARATION} var_decl.parent as struct then
						is_extern := struct.source_info.is_system_source
						flags := flags.bit_or (dependency_graph.struct_member_flag)
					end
				else
					-- should not be possible
				end

				if is_extern then
					flags := flags.bit_or (dependency_graph.external_flag)
				end

				if not dependency_graph.has_item_with_key (node) then
					dependency_graph.insert_item (create {DEPENDENCY_GRAPH_ITEM [AST_TYPE]}.make (node, identifier, flags))
				else
					dependency_graph.update_item (node, identifier, flags)
				end
			elseif attached {AST_CAST_EXPRESSION} node.parent then
				identifier := "temp cast expr"
				if not dependency_graph.has_item_with_key (node) then
					dependency_graph.insert_item (create {DEPENDENCY_GRAPH_ITEM [AST_TYPE]}.make (node, identifier, flags))
				else
					dependency_graph.update_item (node, identifier, flags)
				end
			elseif not dependency_graph.has_item_with_key (node) then
				dependency_graph.insert_default_item (node)
			end
		end

	handle_string_h_function_call (node: AST_FUNCTION_CALL)
		local
			as_stmt: BOOLEAN
			name: STRING
			first_arg, second_arg: LIST [AST_TYPE]
			return_type: AST_TYPE
			returns_string: BOOLEAN
		do
			as_stmt := attached {AST_EXPRESSION_STATEMENT} node.parent
			name := node.identifier_str
			return_type := node.declaration_binding.return_type
			returns_string := is_string_type (return_type)

			first_arg := collect_in_sub_tree (node.parameters.first)
			if node.parameters.count >= 2 then
				second_arg := collect_in_sub_tree (node.parameters.at (2))
			end

			if
				(replaceable_string_h_functions_as_statement.has (name) and then as_stmt) or
				replaceable_string_h_functions_as_expression.has (name) or
				all_args_supportable_string_h_functions.has (name)
			then
				-- any combination of Eiffel string and C string for the arguments is supported.
				-- we don't have to do anything
			elseif
				snd_arg_supportable_string_h_functions.has (name)
			then
				-- only for the second argument an Eiffel string is possible
				-- mark the first argument as bad.
				dependency_graph.mark_all_bad (first_arg, External_function_flag)
			else
				-- no Eiffel strings supported
				dependency_graph.mark_all_bad (first_arg, External_function_flag)
				if second_arg /= Void then
					dependency_graph.mark_all_bad (second_arg, External_function_flag)
				end
			end

			if not as_stmt and then returns_string then
				dependency_graph.mark_bad (return_type, External_function_flag)
			end
		end

	handle_printf_function_call (node: AST_FUNCTION_CALL)
		require
			node.identifier_str /= Void
		local
			helper: PRINTF_HELPER
			target_expr: AST_EXPRESSION
		do
			create helper
			if node.identifier_str.is_equal ("sprintf") then
				node.parameters.start
				target_expr := node.parameters.item
				node.parameters.forth

				if
					not attached {AST_VARIABLE_EXPRESSION} target_expr.non_parentheses_expression as target_var or else
					not helper.is_printf_replaceable (node.parameters)
				then
					-- not supported
					dependency_graph.mark_all_bad (collect_in_sub_tree (target_expr), External_function_flag)
				end
			end
		end
end
